package kimd.VNC.LookOverYourShoulder;
//
//  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

//
// vncviewer.java - the VNC viewer applet.  This class mainly just sets up the
// user interface, leaving it to the vncCanvas to do the actual rendering of
// a VNC desktop.
//

import java.awt.*;
import java.awt.event.*;
import java.awt.event.WindowAdapter;
import java.io.*;
import java.util.*;
import java.lang.*;
import java.net.*;
import java.lang.Object.*;

import netscape.security.PrivilegeManager;
import webwisdom.tango.*;
import netscape.javascript.JSObject;

public class vncviewer extends java.applet.Applet implements TDataListener, TControlListener {
  boolean inAnApplet = true;

  //
  // main() is called when run as a java program from the command line.  It
  // simply creates a frame and runs the applet inside it.
  //

  public static void main(String[] argv) {
    vncviewer v = new vncviewer();
    v.mainArgs = argv;
    v.inAnApplet = false;

    v.f = new Frame("VNC");
    v.f.add("Center", v);

    v.init();
    v.start();
  }

  // TANGO variables
  TAgent agent = null;
  boolean isMaster = true;
  String myName = null;
  String masterName = null;
  String myRole = null;
  boolean authorized = false;

  private int channel = 25;
  private boolean javaOnly = true;
  boolean inTango = true;
  boolean fromScript = false;

  Frame f;
  ScrollPane sp;
  String[] mainArgs;
  String host = null;
  int port;
  String pw;
  String vncControl = null;
  String vncPolling = null;

  rfbProto rfb;
  Thread[] rfbThread;
  VNCviewerThread newThread;
  Vector rfbThreadVector = new Vector();
  int threadIndex = 0;
  String userName = null;

  GridBagLayout gridbag;
  Panel buttonPanel;
  Button disconnectButton;
  Button optionsButton;
  Button clipboardButton;
  Button ctrlAltDelButton;
  optionsFrame options;
  clipboardFrame clipboard;
  authenticationPanel authenticator;
  warningSign ws;

  // message variables
  final String messageSpace	= " ";
  final String messageInitial	= "Initial";
  final String messageData	= "Data";
  final String messageName	= "Name";
  final String messagePort	= "Port";
  final String messageHost	= "Host";
  final String messagePW	= "PW";
  final String messageEnd	= "kimd";

  final String remoteFileName   = "tangoVNC";
  final String tangoVNCname     = "tangoVNC\\tangoVNC";

  boolean connected = false;
  boolean debugApplet = true;

  int framebufferWidth = 0;
  int framebufferHeight = 0;
  String desktopName = null;

  InetAddress localAddr;
  String localHostname = null;
  String tangoVNCdir = " ";
  public String fileName = null;

  public Vector authorPeople = new Vector();
  public Vector userList = new Vector();

  Label title1, title2, retry;
  java.awt.List tangoList;
  Button connect;
  boolean ready2connect = false;

  public void printLine(String printData) {
    if ( debugApplet ) {
      System.out.println("<LOYS> " + printData);
    }
  }
 
  //
  // init()
  //

  public void init() {

    initTango();
    if (inTango) {
      readParameters();
    }
    else {
      fileName="tangoFile";
    }

    try {
      localHostname = java.net.InetAddress.getLocalHost().getHostAddress();
    } 
    catch (UnknownHostException e) {
    }
    printLine("hostname : " + localHostname);

    threadIndex = 0;

    initApplet();
    initThis();
  }

  public void initThis() {

    gridbag = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();

    ready2connect = false;

    if ( authorized ) {
      title1 = new Label("Look Over Your Shoulder",Label.CENTER);
      title1.setFont(new Font("Helvetica", Font.BOLD, 18));

      tangoList = new java.awt.List();
      tangoList.setFont(new Font("Dialog", Font.PLAIN, 14));
      tangoList.addItemListener(new ItemListener() {
	public void itemStateChanged(ItemEvent e) {
	  int selectedItem  = tangoList.getSelectedIndex();
	  printLine("list : " + selectedItem);
	  ready2connect = false;
	  userName = tangoList.getSelectedItem();
	  if ( myName.equals(userName) ) {
	    retry.setText("Can Not Connect to Yourself");
	    return;
	  }
	  retry.setText("");
	  ready2connect = true;

	  port = Integer.parseInt((String)((Vector)userList.elementAt(selectedItem)).elementAt(1));
	  host = (String)((Vector)userList.elementAt(selectedItem)).elementAt(2);
	  pw = (String)((Vector)userList.elementAt(selectedItem)).elementAt(3);
	}
      });

      connect = new Button("Connect");
      connect.setFont(new Font("Courier", Font.BOLD, 14));
      connect.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	  if ( ready2connect ) {
	    newViewThread();
	  }
	}
      });

      retry = new Label("                                   ",Label.CENTER);
      retry.setFont(new Font("Courier", Font.BOLD, 14));

      setLayout(gridbag);

      gbc.gridx = 0;
      gbc.gridy = 0;
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gridbag.setConstraints(title1,gbc);
      add(title1);

      gbc.gridx = 0;
      gbc.gridy = 1;
      gbc.fill = GridBagConstraints.REMAINDER;
      gridbag.setConstraints(tangoList,gbc);
      add(tangoList);

      gbc.gridx = 0;
      gbc.gridy = 2;
      gbc.fill = GridBagConstraints.REMAINDER;
      gridbag.setConstraints(connect,gbc);
      add(connect);

      gbc.gridx = 0;
      gbc.gridy = 3;
      gbc.fill = GridBagConstraints.REMAINDER;
      gridbag.setConstraints(retry,gbc);
      add(retry);
    }
    else { // not authorized person
      printLine("not authorized");
      if ( isMaster ) {
	title1 = new Label("Look Over Your Shoulder",Label.CENTER);
	title2 = new Label("You are not Authorized",Label.CENTER);
      }
      else {
	title1 = new Label("Look Over Your Shoulder",Label.CENTER);
	title2 = new Label("Caution! Your Screen is being Shared",Label.CENTER);
      }
      title1.setFont(new Font("Helvetica", Font.BOLD, 18));
      title2.setFont(new Font("Helvetica", Font.BOLD, 18));

      setLayout(gridbag);
      gbc.gridx = 0;
      gbc.gridy = 0;
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gridbag.setConstraints(title1,gbc);
      add(title1);

      gbc.gridx = 0;
      gbc.gridy = 1;
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gridbag.setConstraints(title2,gbc);
      add(title2);
    }
    validate();
    if ( !authorized &&
	 isMaster ) {
      if (agent != null) {
	agent.exit();
      }
      super.destroy();
      return;
    }

    invokeWinVNC(true);

    options = new optionsFrame(this);
    clipboard = new clipboardFrame(this);
    authenticator = new authenticationPanel(this);
    ws = new warningSign(!authorized, isMaster);

    /**


    if ( !authorized &&
	 isMaster ) {
      gridbag = new GridBagLayout();
      GridBagConstraints gbc = new GridBagConstraints();
      gbc.gridwidth = GridBagConstraints.REMAINDER;
      gbc.anchor = GridBagConstraints.NORTHWEST;
      gbc.weightx = 1.0;
      gbc.weighty = 1.0;
      gbc.ipadx = 100;
      gbc.ipady = 50;
      gridbag.setConstraints(ws,gbc);
      add(ws);
      validate();
      destroy();
      return;
    }
    rfbThread = new Thread(this);
    rfbThread.start();
    **/

  }

  public void listAdded() {
    if ( !authorized ) {
      return;
    }
    tangoList.removeAll();
    // read the list from the vector
    for ( int i = 0 ; i < userList.size() ; i++ ) {
      tangoList.add((String)((Vector)userList.elementAt(i)).firstElement());
    }
  }

  public void newViewThread() {
    printLine("ready for vncviewer");
    newThread = new VNCviewerThread(this, options, userName, port, host, pw);
  }

  public void initTango() {

    String arg = null;
    try {
      arg = getParameter("interface");
      inTango = true;
    }
    catch(Exception e) {
      inTango = false;
    }

    if (arg == null) {
      inTango = false;
    }
    printLine("arg : " + arg);
    printLine("inTango : " + inTango);

    if (inTango) {
      setTango(arg);
    }
  }

  public void initApplet() {
    printLine("inTango  : " + inTango);
    printLine("isMaster : " + isMaster);
    if ( inTango ) {
      host = localHostname;
      myName = agent.getUserName();
      masterName = agent.getMasterName();
      readVNCauthor();
      readVNCconfig();
      //invokeWinVNC(true);

      // slave sends the initial request
      if ( !isMaster ) {
	String tmpMessage = messageInitial + messageSpace
                          + myName         + messageSpace
	                  + port           + messageSpace
	                  + host           + messageSpace
	                  + pw             + messageSpace
                          + messageEnd;
	sendMessage(tmpMessage);
      }
      else {
	Vector newUser = new Vector();
	String name = myName;
	String tmpHost = host;
	String tmpPort = port + messageSpace;
	String tmpPW = pw;

	newUser.addElement(name);
	newUser.addElement(tmpPort);
	newUser.addElement(tmpHost);
	newUser.addElement(tmpPW);

	userList.addElement(newUser);
      }
    }
    else {
      isMaster = true;
    }
  }

  public void setTango(String arg) {
    if ( arg.equals("javaApplet") ) {
      fromScript = false;
      commJava();
    }
    else if ( arg.equals("javaScript") ) {
      fromScript = true;
      commJavaScript();
    }

    while ( agent == null ) {
      printLine("waiting for the agent to be assigned");
    }
    printLine("agent assigned");

    agent.addTDataListener(channel, this);
    agent.addTControlListener(this);
    isMaster = agent.isMaster();

    tangoVNCdir = agent.getApplicationPath();
    tangoVNCdir = tangoVNCdir + tangoVNCname;
    printLine("dir : " + tangoVNCdir);
  }

  public void setTango(String args[]) {
    try {
      agent = new TAgent(new TLAgentApplication(args));
    }
    catch(TangoException e) {
      e.printStackTrace();
      System.out.println("<VNC> agent is not proprely assigned");
    }

    printLine("Application2 agent assigned");

    agent.addTDataListener(channel, this);
    printLine("Data added");

    agent.addTControlListener(this);
    printLine("Control added");

    //isMaster = agent.isMaster();
    printLine("Data Control added");
  }

  public void commJava() {
    try {
      agent = new TAgent(new TLAgentApplet(this));
    }
    catch(TangoException e) {
      e.printStackTrace();
      System.out.println("<VNC> agent is not proprely assigned");
    }
  }

  public void commJavaScript() {
    JSObject jsWin = JSObject.getWindow(this);
    Object[] args = new Object[1];
    args[0] = this;
    jsWin.call("appletReadyJS", args);
  }

  public void initTangoJS(TAgent a, Number ch, Number ma) {
    channel = ((Number) ch).intValue();

    if ( ((Number) ma).intValue() ==  0 ) {
      isMaster = true;
    }
    else {
      isMaster = false;
    }
    agent = a;
  }

  public void channelReceive(int ch, byte b[]) {
    if ( channel != ch ) {
      return;
    }
    receive(b);
  }

  public void masterChangedJS(Number ma, String masterName) {
    if ( ((Number) ma).intValue() ==  0 ) {
      isMaster = true;
    }
    else {
      isMaster = false;
    }

  }

  /*****************************************************************
   *
   *	method	 : masterchanged()
   *	Input    : boolean master, String masterName
   *	Output   : none
   *	Function : Built-in function for TControlListener
   *
   *****************************************************************/
  public void masterChanged(boolean master, String masterName) {
    isMaster = master;
  }

  /*****************************************************************
   *
   *	method	 : participantJoined()
   *	Input    : String participantName
   *	Output   : none
   *	Function : Built-in function for TControlListener
   *
   *****************************************************************/
  public void participantJoined(String participantName) {
  }

  /*****************************************************************
   *
   *	method	 : participantLeft()
   *	Input    : String participantName
   *	Output   : none
   *	Function : Built-in function for TControlListener
   *
   *****************************************************************/
  public void participantLeft(String participantName) {
    if ( !authorized ) {
      return;
    }
    printLine("before remove : " + userList);
    tangoList.remove(participantName);
    for ( int i = 0 ; i < userList.size() ; i++ ) {
      if ( ((Vector)userList.elementAt(i)).contains(participantName) ) {
	userList.removeElementAt(i);
      }
    }
    printLine("after remove : " + userList);
  }

  /*****************************************************************
   *
   *	method	 : audioRequested()
   *	Input    : boolean isREquested
   *	Output   : none
   *	Function : Built-in function for TControlListener
   *
   *****************************************************************/
  public void audioRequested(boolean isREquested) {

  }

  /*****************************************************************
   *
   *	method	 : stringToByte()
   *	Input    : String s
   *	Output   : byte[]
   *	Function : convert String to Byte
   *
   *****************************************************************/
  private byte[] stringToBytes(String s) {
    int l = s.length();
    byte[] t = new byte[l];
    s.getBytes(0, l, t, 0);
    return t;
  }

  /*****************************************************************
   *
   *	method	 : bytesToString()
   *	Input    : byte t[]
   *	Output   : String
   *	Function : convert byte to String
   *
   *****************************************************************/
  private String bytesToString(byte t[]) {
    return new String(t, 0);
  }

  /*****************************************************************
   *
   *	method	 : destroy()
   *	Input    : none
   *	Output   : none
   *	Function : 
   *
   *****************************************************************/
  public void destroy() {
    printLine("inside destroy()");

    if ( fromScript ) {
      agent.removeTControlListener(this);
      agent.removeTDataListener(this);
      printLine("data and control removed");
      return;
    }

    if (agent != null) {
      agent.exit();
    }
    super.destroy();

    removeAll();
    for ( int i = 0 ; i < rfbThreadVector.size() ; i++ ) {
      if ( (Thread)rfbThreadVector.elementAt(i) != null ) {
      }
    }
    if ( rfb != null ) {
      rfb.close();
      rfb = null;
    }
    invokeWinVNC(false);
  }

  /*****************************************************************
   *
   *	method	 : receive()
   *	Input    : byte b[]
   *	Output   : none
   *	Function : built-in function of TDataListener
   *
   *****************************************************************/
  public void receive(byte b[]) {
    String recvMessage = bytesToString(b);
    printLine(" received : " + recvMessage);

    StringTokenizer selectedItem = null;
    selectedItem = new StringTokenizer(recvMessage);
    String head = selectedItem.nextToken();
    if ( isMaster ) {
      if ( head.equals(messageInitial) ) {
	Vector newUser = new Vector();
	String name = selectedItem.nextToken();
	String port = selectedItem.nextToken();
	String host = selectedItem.nextToken();
	String pw = selectedItem.nextToken();

	newUser.addElement(name);
	newUser.addElement(port);
	newUser.addElement(host);
	newUser.addElement(pw);

	userList.addElement(newUser);

	// sends the data - list of host, port, password
	String dataToBeSent = messageData + messageSpace
	                    + userList.size();

	for ( int i = 0 ; i < userList.size() ; i++ ) {
	  dataToBeSent = dataToBeSent + messageSpace
	               + (String) ((Vector)userList.elementAt(i)).elementAt(0) + messageSpace
	               + (String) ((Vector)userList.elementAt(i)).elementAt(1) + messageSpace
	               + (String) ((Vector)userList.elementAt(i)).elementAt(2) + messageSpace
	               + (String) ((Vector)userList.elementAt(i)).elementAt(3);
	}
	dataToBeSent = dataToBeSent + messageSpace
	             + messageEnd;
	sendMessage(dataToBeSent);
	listAdded();
      }
    }
    else { // slave
      if ( head.equals(messageData) ) {
	int noOfUsers = Integer.parseInt(selectedItem.nextToken());
	userList.removeAllElements();
	for ( int i = 0 ; i < noOfUsers ; i++ ) {
	  Vector newUser = new Vector();

	  String name = selectedItem.nextToken();
	  String port = selectedItem.nextToken();
	  String host = selectedItem.nextToken();
	  String pw = selectedItem.nextToken();

	  newUser.addElement(name);
	  newUser.addElement(port);
	  newUser.addElement(host);
	  newUser.addElement(pw);

	  userList.addElement(newUser);
	}
	listAdded();
      }
    }
  }

  /*****************************************************************
   *
   *	method	 : sendMessage()
   *	Input    : String messageToBeSent
   *	Output   : none
   *	Function : send the message and draw it on the Graph
   *
   *****************************************************************/
  void sendMessage(String messageToBeSent) {
    if (inTango) {
      printLine("send : " + messageToBeSent);
      agent.send(channel, stringToBytes(messageToBeSent));
    }
  }

  public void update(Graphics g) {
  }

  public void closeApplet() {
    printLine("closeApplet");
    try {
      JSObject jsWin = JSObject.getWindow(this);
      JSObject topWin = (JSObject)jsWin.getMember("top");
      topWin.call("close", null);
    }
    catch (Exception e) {
      printLine("close error");
    }
  }

  void readVNCauthor() {

    host = localHostname;
    myRole = agent.getUserRole(myName);
    printLine("My Role : " + myRole);

    try {
      PrivilegeManager.enablePrivilege("UniversalConnect");
    }
    catch (netscape.security.ForbiddenTargetException e) {
      System.out.println("\tFailed! Permission to connect.");
    }
    catch (Exception e) {
      System.out.println("\tFailed! Unknown exception while enabling privilege.");
    }
    String fileURL = null;
    String lastToken = null;
    try {
      fileURL = getDocumentBase().toString();
      printLine("DocumentBase : " + fileURL);

      StringTokenizer tmpItem = null;
      tmpItem = new StringTokenizer(fileURL, "/");
      String tmpToken = null;
      while ( true ) {
	tmpToken = tmpItem.nextToken();
	lastToken = tmpToken;
      }
    }
    catch (NoSuchElementException e) {
      printLine("end of token");
    }

    URL url = null;
    BufferedReader in = null;
    String authorFile = null;
    boolean securedServer = false;
    if ( myRole == null ) {
      printLine("Check Role null");
      authorFile = "/AuthorizedPeople";
      securedServer = false;
    }
    else {
      authorFile = "/AuthorizedRole";
      securedServer = true;
    }

    try {
      fileURL = fileURL.substring(0, fileURL.lastIndexOf(lastToken)-1);
      fileURL = fileURL + authorFile;
      printLine("new url : " + fileURL);
      url = new URL(fileURL);
      in = new BufferedReader(new InputStreamReader(url.openStream()));

      String inputLine;
      while ( (inputLine = in.readLine()) != null ) {
	printLine("line : " + inputLine);
	StringTokenizer tmpItem = null;
	tmpItem = new StringTokenizer(inputLine);
	String head = tmpItem.nextToken();

	printLine("head : " + head);
	authorPeople.addElement(head);
      }
    }
    catch (Exception e) {
      printLine("error");
    }
    printLine("authorized People : " + authorPeople);

    if ( securedServer ) {
      if ( authorPeople.contains(myRole) ) {
	authorized = true;
      }
      else {
	authorized = false;
      }
    }
    else {
      if ( authorPeople.contains(myName) ) {
	authorized = true;
      }
      else {
	authorized = false;
      }
    }
  }

  void readVNCconfig() {

    host = localHostname;

    try {
      PrivilegeManager.enablePrivilege("UniversalConnect");
    }
    catch (netscape.security.ForbiddenTargetException e) {
      System.out.println("\tFailed! Permission to connect.");
    }
    catch (Exception e) {
      System.out.println("\tFailed! Unknown exception while enabling privilege.");
    }
    String fileURL = null;
    String lastToken = null;
    String last_1Token = null;
    try {
      fileURL = getDocumentBase().toString();
      printLine("DocumentBase : " + fileURL);

      StringTokenizer tmpItem = null;
      tmpItem = new StringTokenizer(fileURL, "/");
      String tmpToken = null;
      while ( true ) {
	tmpToken = tmpItem.nextToken();
	last_1Token = lastToken;
	lastToken = tmpToken;
      }
    }
    catch (NoSuchElementException e) {
      printLine("end of token");
    }

    URL url = null;
    BufferedReader in = null;
    try {
      fileURL = fileURL.substring(0, fileURL.lastIndexOf(lastToken)-1);
      fileURL = fileURL + "/tangoVNC";
      printLine("new url : " + fileURL);
      url = new URL(fileURL);
      in = new BufferedReader(new InputStreamReader(url.openStream()));

      String inputLine;
      while ( (inputLine = in.readLine()) != null ) {
	printLine("line : " + inputLine);
	StringTokenizer tmpItem = null;
	tmpItem = new StringTokenizer(inputLine);
	String head = tmpItem.nextToken();
	String data = tmpItem.nextToken();

	printLine("head : " + head + "     data : " + data);

	if ( head.equalsIgnoreCase("port") ) {
	  port = Integer.parseInt(data);
	}
	else if ( head.equalsIgnoreCase("password") ) {
	  pw = data;
	}
	else if ( head.equalsIgnoreCase("control") ) {
	  vncControl = data;
	}
	else if ( head.equalsIgnoreCase("polling") ) {
	  vncPolling = data;
	}
      }
    }
    catch (Exception e) {
      printLine("error");
    }
  }

  void invokeWinVNC(boolean action) {
    // invoke the vnc server
    try {
      PrivilegeManager.enablePrivilege("UniversalExecAccess");
    }
    catch (netscape.security.ForbiddenTargetException e) {
      System.out.println("\tFailed! Permission to connect.");
    }
    catch (Exception e) {
      System.out.println("\tFailed! Unknown exception while enabling privilege.");
    }

    String[] cmdarray = new String[1];
    if ( action ) {
      cmdarray[0] = tangoVNCdir
	          + " -LookOverYourShoulder"
	          + " -port " + port
	          + " -password " + pw;
      if ( vncControl.equals("fullcontrol") ) {
	cmdarray[0] = cmdarray[0] +" -fullcontrol";
      }
      else if ( vncControl.equals("monitor") ) {
	cmdarray[0] = cmdarray[0] + " -monitor";
      }
      if ( vncPolling.equals("fullscreen") ) {
	cmdarray[0] = cmdarray[0] + " -fullscreen";
      }
      else if ( vncPolling.equals("foreground") ) {
	cmdarray[0] = cmdarray[0] + " -foreground";
      }
      else if ( vncPolling.equals("cursor") ) {
	cmdarray[0] = cmdarray[0] + " -cursor";
      }
    }
    else {
      cmdarray[0] = tangoVNCdir
	          + " -LookOverYourShoulder -kill";
      /**
      cmdarray[1] = "-kill";
      cmdarray[2] = null;
      cmdarray[3] = null;
      cmdarray[4] = null;
      **/
    }
    printLine("command : " + cmdarray[0]);
    try {
      Runtime.getRuntime().exec(cmdarray);
    }
    catch (IOException e) {
      printLine("not working");
    }
  }

  //
  // Do the rest of the protocol initialisation.
  //

  void doProtocolInitialisation() throws IOException {
    rfb.writeClientInit();

    rfb.readServerInit();

    desktopName = rfb.desktopName;
    framebufferWidth  = rfb.framebufferWidth;
    framebufferHeight = rfb.framebufferHeight;

    printLine("Desktop name is " + desktopName);
    printLine("Desktop size is " + framebufferWidth + " * " +
		       framebufferHeight);

    setEncodings();
  }


  //
  // setEncodings() - send the current encodings from the options frame
  // to the RFB server.
  //

  void setEncodings() {
    try {
      if ((rfb != null) && rfb.inNormalProtocol) {
	rfb.writeSetEncodings(options.encodings, options.nEncodings);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }


  //
  // setCutText() - send the given cut text to the RFB server.
  //

  void setCutText(String text) {
    try {
      if ((rfb != null) && rfb.inNormalProtocol) {
	rfb.writeClientCutText(text);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }


  //
  // Respond to an action i.e. button press
  //

  public synchronized boolean action(Event evt, Object what) {

    if (evt.target == optionsButton) {

      if (options.isVisible()) {
	options.hide();
      } else {
	options.show();
      }

    } else if (evt.target == disconnectButton) {

      printLine("disconnect");
      options.dispose();
      clipboard.dispose();

      if (inAnApplet) {
	removeAll();
	rfb.close();
	rfb = null;
	Label l = new Label("Disconnected");
	setLayout(new FlowLayout(FlowLayout.LEFT, 30, 30));
	add(l);
	validate();
	rfbThread[threadIndex].stop();
      } else {
	System.exit(1);
      }

    } else if (evt.target == clipboardButton) {

      if (clipboard.isVisible()) {
	clipboard.hide();
      } else {
	clipboard.show();
      }

    } else if (evt.target == ctrlAltDelButton) {

      try {
	Event ctrlAltDelEvent = new Event(null, 0, null);

	ctrlAltDelEvent.key = 127;
	ctrlAltDelEvent.modifiers = Event.CTRL_MASK | Event.ALT_MASK;

	ctrlAltDelEvent.id = Event.KEY_PRESS;
	rfb.writeKeyEvent(ctrlAltDelEvent);

	ctrlAltDelEvent.id = Event.KEY_RELEASE;
	rfb.writeKeyEvent(ctrlAltDelEvent);
      } catch (Exception e) {
	e.printStackTrace();
      }
    }
    return false;
  }


  //
  // Detect when the focus goes in and out of the applet.  See
  // vncCanvas.handleEvent() for details of why this is necessary.
  //

  boolean gotFocus = false;

  public boolean gotFocus(Event evt, Object what) {
    gotFocus = true;
    return true;
  }
  public boolean lostFocus(Event evt, Object what) {
    gotFocus = false;
    return true;
  }


  //
  // encryptBytes() - encrypt some bytes in memory using a password.  Note that
  // the mapping from password to key must be the same as that used on the rfb
  // server side.
  //
  // Note also that IDEA encrypts data in 8-byte blocks, so here we will ignore
  // any data beyond the last 8-byte boundary leaving it to the calling
  // function to pad the data appropriately.
  //

  void encryptBytes(byte[] bytes, String passwd) {
    byte[] key = new byte[8];
    passwd.getBytes(0, passwd.length(), key, 0);

    for (int i = passwd.length(); i < 8; i++) {
      key[i] = (byte)0;
    }

    DesCipher des = new DesCipher(key);

    des.encrypt(bytes,0,bytes,0);
    des.encrypt(bytes,8,bytes,8);
  }


  //
  // readParameters() - read parameters from the html source or from the
  // command line.  On the command line, the arguments are just a sequence of
  // param_name/param_value pairs where the names and values correspond to
  // those expected in the html applet tag source.
  //

  public void readParameters() {
    if ( !isMaster ) {
      return;
    }

    fileName = readParameter("FILE", false);
    /**
    host = readParameter("HOST", !inAnApplet);
    if (host == null) {
      host = getCodeBase().getHost();
      if (host.equals("")) {
	fatalError("HOST parameter not specified");
      }
    }

    String s = readParameter("PORT", true);
    port = Integer.parseInt(s);

    pw = readParameter("PASS", true);
    **/
  }

  public String readParameter(String name, boolean required) {
    if (inAnApplet) {
      String s = getParameter(name);
      if ((s == null) && required) {
	fatalError(name + " parameter not specified");
      }
      return s;
    }

    for (int i = 0; i < mainArgs.length; i += 2) {
      if (mainArgs[i].equalsIgnoreCase(name)) {
	try {
	  return mainArgs[i+1];
	} catch (Exception e) {
	  if (required) {
	    fatalError(name + " parameter not specified");
	  }
	  return null;
	}
      }
    }
    if (required) {
      fatalError(name + " parameter not specified");
    }
    return null;
  }

  //
  // fatalError() - print out a fatal error message.
  //

  public void fatalError(String s) {
    printLine(s);

    StringTokenizer tmpToken = new StringTokenizer(s);
    String tmpHead = tmpToken.nextToken();
    String errorStr = s.substring(tmpHead.length()+1);

    if (inAnApplet) {
      removeAll();
      //Label l = new Label(s);
      Label l = new Label(errorStr,Label.CENTER);
      l.setFont(new Font("Helvetica", Font.BOLD, 18));

      setLayout(new FlowLayout(FlowLayout.LEFT, 30, 30));
      add(l);
      validate();
      Thread.currentThread().stop();
    } else {
      System.exit(1);
    }

    /**
    if ( fromScript ) {
      agent.removeTControlListener(this);
      agent.removeTDataListener(this);
      printLine("data and control removed");
      return;
    }

    if (agent != null) {
      agent.exit();
    }
    **/
    destroy();
  }
}
